@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter ();
 
-our $VERSION = '0.012';
+our $VERSION = '0.013';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 262144;
@@ -161,6 +161,8 @@ my %HTTP_STATUS = (
 sub _handle_error {
   my ($cgi, $error) = @_;
   return unless $cgi->{pid} == $$; # in case of fork
+  $cgi->{response_status} = "500 $HTTP_STATUS{500}" unless $cgi->{headers_rendered}
+    or (defined $cgi->{response_status} and $cgi->{response_status} =~ m/^[45][0-9]{2} /);
   if (defined(my $handler = $cgi->{on_error})) {
     my ($error_error, $error_errored);
     {
@@ -175,10 +177,7 @@ sub _handle_error {
   } else {
     warn $error;
   }
-  unless ($cgi->{headers_rendered}) {
-    $cgi->{response_status} = "500 $HTTP_STATUS{500}" unless defined $cgi->{response_status};
-    $cgi->set_response_type('text/plain')->render(data => $cgi->{response_status});
-  }
+  $cgi->set_response_type('text/plain')->render(data => $cgi->{response_status}) unless $cgi->{headers_rendered};
 }
 
 sub set_error_handler          { $_[0]{on_error} = $_[1]; $_[0] }
@@ -706,7 +705,8 @@ sub _response_headers {
     $headers_str = "Date: $date_str\r\n$headers_str";
   }
   my $status = $self->{response_status};
-  $status = $self->{response_status} = "302 $HTTP_STATUS{302}" if !defined $status and $type eq 'redirect';
+  $status = $self->{response_status} = "302 $HTTP_STATUS{302}" if $type eq 'redirect'
+    and !(defined $status and $status =~ m/^3[0-9]{2} /);
   if ($self->{nph}) {
     $status = "200 $HTTP_STATUS{200}" unless defined $status;
     my $protocol = $ENV{SERVER_PROTOCOL};
